#ifndef SERIALIZABLE_H
#define SERIALIZABLE_H

#include <unordered_map>
#include "property.h"

/**
// To use these macros, your class should inherit the abstract class "Serializable" :

class MyClass : public Serializable

// you should first declare properties like this :

private:
    INT(myIntProperty)
    REFERENCE(myPointer)
    VECTOR_FLOAT(myFloatArray)

// and then, you'll be able to automatically save these properties after using the following macro :

public:
    SERIALIZABLE(MyClass, CAST(myIntProperty), CAST(myPointer), CAST(myFloatArray))

// to be able to load your saved object, you will need to add the following declaration in the CPP file of your class :

INIT_SERIALIZABLE(MyClass)

// that's it ! now you can use the ObjectSaver and ObjectLoader classes to serialize your objects automatically.
*/

// ALTERNATIVE INITIALIZATION :

// use this if you really can't have a cpp file for your class
#define MANUAL_SERIALIZABLE(className, ...)\
    virtual const std::string &getType() const { return className::getStaticType(); }\
    static const std::string &getStaticType() { static const std::string name = #className; return name; }\
    virtual PropertySet* getProperties() const { return initProperties(new PropertySet(), __VA_ARGS__); }\
    static Serializable* instantiate() { return (Serializable*)(new className()); }\

// can be called to init the class manually whenever you want in your program (before any deserialization of this class)
#define MANUAL_INIT_SERIALIZABLE(className) Serializable::addType(#className, className::instantiate);

// RECOMMENDED INITIALIZATION

#define SERIALIZABLE(className, ...)\
    virtual const std::string &getType() const { return className::getStaticType(); }\
    static const std::string &getStaticType() { static const std::string name = #className; return name; }\
    virtual PropertySet* getProperties() const { return initProperties(new PropertySet(), __VA_ARGS__); }\
    static bool init;\
    static Serializable* instantiate() { return (Serializable*)(new className()); }\

// used to auto-init the class at startup (with global variables initialization)
#define INIT_SERIALIZABLE(className) bool className::init = MANUAL_INIT_SERIALIZABLE(className)\

// END OF MACROS

/**
 * @brief Most of the methods and members of this class are only used by the macros and the automatic serializer, you shouldn't have to worry about them.
 */
class Serializable
{
public:
    virtual const std::string &getType() const = 0;
    virtual PropertySet* getProperties() const = 0;

    typedef Serializable* (*Instancer)(void);
    static Instancer getInstancer(const std::string &type) { return (*s_typeMap)[type]; }
private:
    static std::unordered_map<std::string, Instancer> *s_typeMap;
protected:
    static bool addType(const std::string &name, Instancer fun);

    PropertySet* initProperties(PropertySet* properties, AbstractProperty* p) const;

    template <typename... Args> // recursive variadic
    PropertySet* initProperties(PropertySet* properties, AbstractProperty* p, Args... args) const
    {
        properties->push_back(p);
        return initProperties(properties, args...);
    }
};

#endif // SERIALIZABLE_H
